import requests
from . import config
from .exceptions import SerialAPIException

class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

class APIClient(metaclass=SingletonMeta):
    def __init__(self):
        # The init method should only be called once.
        if not config.api_key or not config.base_url:
            raise ValueError("API client requires an API key and base URL")

        self.session = requests.Session()
        self.update_headers()

    def update_headers(self):
        if config.api_key:
            self.session.headers.update({'Authorization': f'Bearer {config.api_key}'})

    def make_api_request(self, endpoint, method, params=None, data=None, files=None):
        self.update_headers()
        if method == "GET":
            response = self._get(endpoint, params)
        elif method == "POST" and not files:
            response = self._post(endpoint, data=data)
        elif method == "POST" and files:
            response = self._post_files(endpoint, files=files)
        elif method == "PUT":
            response = self._put(endpoint, data)
        elif method == "PATCH":
            response = self._patch(endpoint, data)
        elif method == "DELETE":
            response = self._delete(endpoint, data)
        else:
            print(f"{method} is not supported by the serial python library")
            return None
        
        if not response.ok:
            raise SerialAPIException(f"API request failed: {response.text}", status_code=response.status_code)
        return response.json()
    
    def _error(self, message):
        print(f"ERROR: {message}")

    def _log(self, message):
        pass
        # TODO: debug logging
        #print(f"DEBUG: {message}")

    def _get(self, endpoint, params=None):
        self._log(f"GET request to {endpoint} with params {params}")
        response = self.session.get(f"{config.base_url}{endpoint}", params=params) 
        return response

    def _post(self, endpoint, data=None):
        self._log(f"POST request to {endpoint} with data {data}")
        response = self.session.post(f"{config.base_url}{endpoint}", json=data) 
        return response

    def _post_files(self, endpoint, files):
        self._log(f"POST request to {endpoint} with file") 
        response = self.session.post(f"{config.base_url}{endpoint}", files=files)
        return response

    def _put(self, endpoint, data=None):
        self._log(f"PUT request to {endpoint} with data {data}")
        response = self.session.put(f"{config.base_url}{endpoint}", json=data)
        return response

    def _patch(self, endpoint, data=None):
        self._log(f"PATCH request to {endpoint} with data {data}")
        response = self.session.patch(f"{config.base_url}{endpoint}", json=data)
        return response

    def _delete(self, endpoint, data=None):
        self._log(f"DELETE request to {endpoint} with data {data}")
        response = self.session.delete(f"{config.base_url}{endpoint}", json=data)
        return response




"""
This file contains the ComponentInstanceLink class, which is used to represent
a component instance link
"""
class ComponentInstanceLink:
    """
    A component instance link object, as defined at
    https://docs.serial.io/api-reference/component-instance-links/get-component-instance-link
    """
    def __init__(self, link_data):
        """
        Args:
        - link_data: A component instance link object, as defined at
        https://docs.serial.io/api-reference/component-instance-links/get-component-instance-link

        Returns:
        - A component instance link object (available at data), as ned at
        https://docs.serial.io/api-reference/component-instance-links/get-component-instance-link
        """
        self.data = link_data



"""
This file contains the Dataset class and the Datasets class.
"""
from ..api_client import APIClient
from ..exceptions import SerialAPIException
class DatasetNotFound(SerialAPIException):
    """
    Exception for when a dataset is not found
    """
    pass

class Datasets:
    """
    A class for dataset data methods
    """
    @staticmethod
    def get(name, data_type, process_id):
        """
        Gets a dataset, if it exists
        
        Args:
        - name: Dataset name
        - data_type: Dataset data type

        Returns:
        - A dataset Python object 
        """
        client = APIClient()
        query_params = {"name": name, "type": data_type, "process_id": process_id}
        # TODO: debug logging
        #print(f"Getting dataset {name} of type {data_type}")

        dataset_data = client.make_api_request("/datasets", "GET", params=query_params)
        if dataset_data and len(dataset_data) > 0:
            return Dataset(dataset_data[0])
        else:
            raise DatasetNotFound(f"Dataset {name} of type {data_type} not found")

    
    @staticmethod
    def create(name, data_type, process_id, extra_params=None):
        """
        Creates a dataset, if it does not exist
        
        Args:
        - name: Dataset name
        - data_type: Dataset data type
        - process_id: Process ID
        - extra_params?: Extra parameters. Valid options are: usl, lsl, & unit

        Returns:
        - A dataset Python object 
        """
        client = APIClient()
        query_params = {"name": name, "type": data_type, "process_id": process_id}
        if extra_params:
            query_params.update(extra_params)
        # TODO: debug logging
        #print(f"Creating dataset with type {data_type}: {name}")
        return Dataset(client.make_api_request("/datasets", "PUT", data=query_params))

    @staticmethod
    def get_or_create_dataset(name, data_type, process_id, extra_params=None):
        """
        Gets a dataset, if it exists, otherwise creates it
        
        Args:
        - name: Dataset name
        - data_type?: Dataset data type
        - process_id?: Process ID
        - extra_params?: Extra parameters. Valid options are: usl, lsl, & unit

        Returns:
        - A dataset Python object 
        """
        try:
            return Datasets.get(name, data_type, process_id)
        except DatasetNotFound as e:
            return Datasets.create(name, data_type, process_id, extra_params)

class Dataset:
    """
    A dataset object python object
    """
    def __init__(self, dataset_data):
        """
        Args:
        - dataset_data: A dataset object, as defined at
        https://docs.serial.io/api-reference/datasets/get-dataset
        
        Returns:
        - A dataset Python object, which holds the api object at data, the dataset name at name, the dataset id at dataset_id and the (field) id at id
        """
        self.data = dataset_data
        self.name = dataset_data["name"]
        self.dataset_id = dataset_data["id"]




"""serialmfg Module
Exports 3 data classes (they are not initialized, and simply provide 
users with ways to access underlying data structures)
ComponentInstances
ProcessEntries
Datasets

Exports 3 deprecated classes
Serial (for managing connections)
Identifier (for creating component instances)
Process (for creating process entries)
"""
from .identifier import Identifier
from .serial import Serial 
from .process import Process
from .serial_resources.component_instance import ComponentInstances
from .serial_resources.process_entry import ProcessEntries
from .serial_resources.dataset import Datasets
from . import config
from .exceptions import SerialAPIException

def set_api_key(key):
    config.api_key = key

def set_base_url(url):
    config.base_url = url

def set_station_id(id):
    config.station_id = id

# Export the functions so that they can be used as serial.set_api_key, etc.
__all__ = ['set_api_key', 'set_base_url', 'set_station_id', 'SerialAPIException', ...]



"""
DEPRECATED, WILL BE REMOVED IN FUTURE RELEASE
"""
import json
import os
import mimetypes
from warnings import warn

class Process:
    """
    A class representing the process data to be uploaded to the API.
    """
    def __init__(self, identifier, process_id):
        """
        DEPRECATED, WILL BE REMOVED IN A FUTURE RELEASE
        Constructor for the ProcessData class.
        
        Args:
            identifier (str): The identifier for the process data.
        """
        self.process_id = process_id
        self.links = []
        self.parameters = []
        self.images = []
        self.files = []
        self.is_pass = None
        self.identifier = identifier
        warn('''
        The "Process" object is being deprecated and will be removed in future releases.
        Please begin migrating to the new interface provided from `from serialmfg import client`.
        ''')

    def add_link(self, identifier):
        """
        Adds a linked component to the process data.
        
        Args:
            identifier (str): The identifier for the linked component.
        """
        self.links.append({
            "identifier": identifier,
        })

    def add_parameter(self, key_name, value, usl=None, lsl=None, unit=None):
        """
        Adds a parameter to the process data.
        
        Args:
            key_name (str): The key name of the parameter.
            usl (float): The upper specification limit of the parameter.
            lsl (float): The lower specification limit of the parameter.
            value (float): The value of the parameter.
            unit (str): The unit of measurement for the parameter.
        """

        # set data_type to INTEGER, NUMBER, STRING based on the type of value:
        data_type_py = type(value)

        if data_type_py in [int, float]:
            data_type = "NUMBER" if data_type_py == float else "INTEGER"
        elif data_type_py == str:
            data_type = "STRING"
            if usl is not None or lsl is not None:
                raise ValueError("usl and lsl can only be used if value is an integer or float")
        else:
            raise ValueError("data_type must be 'INTEGER', 'NUMBER', or 'STRING'")

        self.parameters.append({
            "key_name": key_name,
            "data_type": data_type,
            "usl": usl, 
            "lsl": lsl,
            "value": value,
            "unit": unit
        })

    def add_image(self, key_name, path):
        """
        Adds an image to the process data.
        
        Args:
            key_name (str): The key name of the image.
            path (str): The path to the image file.
        """
        with open(path, 'rb') as file:
            image_bytes = file.read()
            file_name = os.path.basename(file.name)
            mime_type, _ = mimetypes.guess_type(file_name)
            self.images.append({
                "key_name": key_name,
                "file_name": file_name,
                "bytes": image_bytes,
                "mime_type": mime_type
            })

    def add_file(self, key_name, path):
        """
        Adds a file to the process data.
        
        Args:
            key_name (str): The key name of the file.
            path (str): The path to the image file.
        """
        with open(path, 'rb') as file:
            file_bytes = file.read()
            file_name = os.path.basename(file.name)
            mime_type, _ = mimetypes.guess_type(file_name)
            self.files.append({
                "key_name": key_name,
                "file_name": file_name,
                "bytes": file_bytes,
                "mime_type": mime_type
            })

    def set_pass_fail(self, is_pass):
        """
        Sets the pass status of the process data.
        
        Args:
            is_pass (bool): The pass status of the process data.
        """
        self.is_pass = is_pass

    def build_payload(self, station_id):
        """
        Builds the multipart form data payload for the endpoint
        """
        serial_payload = {
            "identifier": self.identifier,
            "process_id": self.process_id,
            "station_id": station_id,
            "links": self.links,
            "parameters": self.parameters,
            "images": [],
            "files": [], 
            "is_pass": self.is_pass
        }

        for image in self.images:
            serial_payload["images"].append({
                "key_name": image["key_name"],
                "file_name": image["file_name"]
            })

        for file in self.files:
            serial_payload["files"].append({
                "key_name": file["key_name"],
                "file_name": file["file_name"]
            })

        payload = {
            "serial_payload": ("serial_payload.json", json.dumps(serial_payload), "application/json")
        }
        
        for image in self.images:
            mime_type = image["mime_type"]
            if mime_type is None:
                raise ValueError(f"Could not determine MIME type for image: {image['file_name']}")
            payload[image["file_name"]] = (image["file_name"], image["bytes"], mime_type)

        for file in self.files:
            mime_type = file["mime_type"]
            if mime_type is None:
                raise ValueError(f"Could not determine MIME type for file: {file['file_name']}")
            payload[file["file_name"]] = (file["file_name"], file["bytes"], mime_type)

        return payload



"""
DEPRECATED, WILL BE REMOVED IN FUTURE RELEASE
"""
import requests
from warnings import warn

BASE_URL = "https://api.serial.io"


class Serial:
    def __init__(self, api_key, station_id=None, base_url=BASE_URL):
        """
        Initializes a new Station object.

        Args:
        - api_key: The API key for authentication.
        - station_id: The ID of the station.

        Returns:
        - A new Station object.
        """
        self.api_key = api_key
        self.station_id = station_id
        self.base_url = base_url
        warn('''
        The "Serial" object is being deprecated and will be removed in future releases.
        Please begin migrating to the new interface provided from `from serialmfg import client`.
        ''')


    def set_station_id(self, station_id):
        """
        Sets the station ID for the Station object.

        Args:
        - station_id: The ID of the station.
        """
        self.station_id = station_id

    def upload_process_data(self, process):
        """
        Uploads process data to the server.

        Args:
        - process: A ProcessData object containing the data to be uploaded.

        Returns:
        - A requests.Response object representing the server's response.
        """

        endpoint = self.base_url + "/process-data"
        headers = {"Authorization": f"Bearer {self.api_key}"}
        payload = process.build_payload(station_id=self.station_id)
        response = requests.post(endpoint, headers=headers, files=payload)
        return response
    
    def initialize_identifier(self, identifier_object):
        """
        Initializes the identifier on the server.

        Returns:
        - A requests.Response object representing the server's response.
        """
        endpoint = self.base_url + "/identifier"
        headers = {"Authorization": f"Bearer {self.api_key}", "Content-Type": "application/json"}
        payload = {
            "identifier": identifier_object.identifier,
            "component": identifier_object.component,
            "part_number": identifier_object.part_number,
        }
        response = requests.post(endpoint, headers=headers, json=payload)
        return response
    
    def check_connection(self):
        """
        Checks to see if the server is active and if the station ID is valid

        Returns:
        - A requests.Response object representing the server's response.
        """
        endpoint = self.base_url + "/connection"
        headers = {"Authorization": f"Bearer {self.api_key}", "Content-Type": "application/json"}
        payload = {
            "station_id": self.station_id
        }
        response = requests.post(endpoint, headers=headers, json=payload)
        return response

    def station_id(self):
        return self.station_id

    def api_key(self):
        return self.api_key





class SerialAPIException(Exception):
    """Exception raised for errors in the Serial API call.

    Attributes:
        message -- explanation of the error
        status_code -- HTTP status code from the response
    """

    def __init__(self, message, status_code=None):
        self.message = message
        self.status_code = status_code
        super().__init__(self.message)
    
    def __str__(self):
        if self.status_code:
            return f"APIException {self.status_code}: {self.message}"
        else:
            return f"APIException: {self.message}"



"""
This module contains the ProcessEntry class and ProcessEntries class.
"""
import os
import mimetypes
from datetime import datetime
from ..api_client import APIClient 
from ..exceptions import SerialAPIException
from .. import config
from .dataset import Datasets 
from .component_instance import ComponentInstance, ComponentInstances
from .component_instance_link import ComponentInstanceLink

class ProcessEntry:
    """
    A process entry Python object
    """
    def __init__(self, process_entry_data): 
        """
        Creates a new Process Entry
        
        Args:
        - process_entry_data: A process entry object, as defined at
        https://docs.serial.io/api-reference/process-entries/get-process-entry

        Returns:
        - A newly created process entry Python object, which holds the api object at data, the process id at process_id and the id at id
        """
        # TODO: debug logging
        #print(f"Creating process entry object with data: {process_entry_data}")
        self.client = APIClient() 
        self.data = process_entry_data
        self.process_id = process_entry_data["process_id"] 
        self.id = process_entry_data["id"] 
        self.component_instance_id = process_entry_data["unique_identifier_id"]
        self.text_data_queue = []
        self.numerical_data_queue = []
        self.file_data_queue = []
        self.boolean_data_queue = []
        self.link_data_queue = []

    def add_text(self, dataset_name, value, expected_value=None):
        """
        Store text data to be added to a process entry

        Args: 
        - dataset_name: User facing name of the dataset
        - value: The value to be submitted
        - expected_value: Optional argument to pass an expected value. If the 
          values are the same, it would be considered a passing test result

        Stores:
        - Information necessary to add text data
        """
        # Store the necessary information in an internal queue or list
        self.text_data_queue.append({
            'dataset_name': dataset_name,
            'value': value,
            'expected_value': expected_value
        })

    def _process_text_data_queue(self):
        """
        Execute the stored requests to add text data to process entries.
        """
        for text_data in self.text_data_queue:
            dataset_name = text_data['dataset_name']
            value = text_data['value']
            expected_value = text_data.get('expected_value')

            dataset = Datasets.get_or_create_dataset(dataset_name, "TEXT", self.process_id)

            data = {
                "type": "TEXT",
                "dataset_id": dataset.dataset_id,
                "value": value,
                # Include expected_value in the data if provided
                **({'expected_value': expected_value} if expected_value is not None else {})
            }
            
            # Assuming make_api_request can handle the 'data' as it is.
            response = self.client.make_api_request(
                f"/processes/entries/{self.id}", "PUT", data=data
            )

        # Clear the queue once all text data has been added
        self.text_data_queue.clear()

    def add_number(self, dataset_name, value, usl=None, lsl=None, unit=None):
        """
        Queue numerical data to be added to a process entry.

        Args: 
        - dataset_name: User facing name of the dataset
        - value: The value to be submitted
        - usl: Optional argument to override the dataset's upper spec limit
        - lsl: Optional argument to override the dataset's lower spec limit
        - unit: Optional argument to create the dataset's unit
        """
        # Append a task to the queue with all necessary information
        self.numerical_data_queue.append({
            'dataset_name': dataset_name,
            'value': value,
            'usl': usl,
            'lsl': lsl,
            'unit': unit
        })

    def _process_numerical_data_queue(self):
        """
        Process the queued numerical data submissions.
        """
        for number_data in self.numerical_data_queue:
            dataset_name = number_data['dataset_name']
            value = number_data['value']
            usl = number_data.get('usl')
            lsl = number_data.get('lsl')
            unit = number_data.get('unit')
            extra_params = {'unit': unit, 'usl': usl, 'lsl': lsl}
            
            dataset = Datasets.get_or_create_dataset(dataset_name, "NUMERICAL", self.process_id, extra_params)
            
            data = {
                "type": "NUMERICAL",
                "dataset_id": dataset.dataset_id,
                "value": value
            }
            if usl is not None:
                data["usl"] = usl
            if lsl is not None:
                data["lsl"] = lsl

            # Here, make_api_request would be an asynchronous method call
            response = self.client.make_api_request(f"/processes/entries/{self.id}", "PUT", data=data)

    def add_image(self, dataset_name, path, file_name=None):
        """
        DEPRECATED: Use add_file instead

        Args: 
        - dataset_name: User facing name of the dataset
        - path: Path to the file on your file system
        - file_name: Optional argument to override the file name
        """
        return self.add_file(dataset_name, path, file_name)

    def add_file(self, dataset_name, path, file_name=None):
        """
        Queue file data to be added to a process entry.

        Args: 
        - dataset_name: User facing name of the dataset
        - path: Path to the file on your file system
        - file_name: Optional argument to override the file name
        """
        # Queue the task with all necessary information
        self.file_data_queue.append({
            'dataset_name': dataset_name,
            'path': path,
            'file_name': file_name
        })

    def _process_file_data_queue(self):
        """
        Process the queued file data submissions.
        """
        for file_data in self.file_data_queue:
            dataset_name = file_data['dataset_name']
            path = file_data['path']
            file_name = file_data.get('file_name') or os.path.basename(path)

            # This method should be updated to handle the actual file upload and dataset retrieval/creation
            response = self._upload_file(dataset_name, path, file_name, "FILE")
            # Handle the response as necessary

        # Clear the queue once all files have been uploaded
        self.file_data_queue.clear()
    
    def _upload_file(self, dataset_name, path, file_name, dataset_type):
        """
        Uploads a file to the serial server

        Args:
        - dataset_name: User facing name of the dataset
        - path: Path to the file on your file system
        - file_name: None-able argument to override the file name
        - dataset_type: Type of dataset to be uploaded (IMAGE or FILE)
        """
        if not file_name:
            file_name = os.path.basename(path)
        # if in the future we need to do a better guess, we can use python-magic to determine the mimetype
        files = {'file': (file_name, open(path, 'rb'), mimetypes.guess_type(path))} 
        mimetype = mimetypes.guess_type(path)[0]
        storage_object = self.client.make_api_request("/files", "POST", files=files)
        if mimetype.startswith("image"):
            dataset_type = "IMAGE"
        dataset = Datasets.get_or_create_dataset(dataset_name, dataset_type, self.process_id) 
        data = {"type": dataset_type, "dataset_id": dataset.dataset_id, "file_id": storage_object["name"], "file_name": file_name}
        return self.client.make_api_request(f"/processes/entries/{self.id}", "PUT", data=data)

    def add_boolean(self, dataset_name, value, expected_value):
        """
        Queue boolean data to be added to a process entry.

        Args: 
        - dataset_name: User facing name of the dataset
        - value: The value to be submitted
        - expected_value: Argument to pass an expected value.
        """
        # Append a task with all necessary information to the queue
        self.boolean_data_queue.append({
            'dataset_name': dataset_name,
            'value': value,
            'expected_value': expected_value
        })

    def _process_boolean_data_queue(self):
        """
        Process the queued boolean data submissions.
        """
        for boolean_data in self.boolean_data_queue:
            dataset_name = boolean_data['dataset_name']
            value = boolean_data['value']
            expected_value = boolean_data['expected_value']

            # Retrieve or create the dataset as necessary
            dataset = Datasets.get_or_create_dataset(dataset_name, "BOOLEAN", self.process_id)
            
            data = {
                "type": "BOOLEAN",
                "dataset_id": dataset.dataset_id,
                "value": value,
                "expected_value": expected_value
            }

            # Assuming make_api_request can handle the 'data' as it is
            response = self.client.make_api_request(
                f"/processes/entries/{self.id}", "PUT", data=data
            )

            # Handle the response as necessary

        # Clear the queue once all boolean data has been added
        self.boolean_data_queue.clear()

    def add_link(self, dataset_name, child_identifier, break_prior_links=False):
        """
        Queue a link creation between a parent and a child at this specific process entry.

        Args: 
        - dataset_name: User facing name for the link
        - child_identifier: Identifier for the child component instance to be linked
        - break_prior_links: Boolean for whether to break prior links
        """
        # Queue the task with all necessary information
        self.link_data_queue.append({
            'dataset_name': dataset_name,
            'child_identifier': child_identifier,
            'break_prior_links': break_prior_links
        })
    

    def _process_link_data_queue(self):
        """
        Process the queued link creation tasks.
        """
        for link_data in self.link_data_queue:
            dataset_name = link_data['dataset_name']
            child_identifier = link_data['child_identifier']
            break_prior_links = link_data['break_prior_links']

            # Logic to retrieve the child component instance and link dataset
            # Assume `get_child_component_instance` and `get_link_dataset` methods are implemented
            child_component_instance_id = ComponentInstances.get(child_identifier).data["id"]
            link_dataset = Datasets.get(dataset_name, "LINK", self.process_id)

            # Now, create the link using the retrieved information
            process_entry_id = self.id
            data = {
                "parent_component_instance_id": self.component_instance_id, 
                "child_component_instance_id": child_component_instance_id,
                "dataset_id": link_dataset.dataset_id,
                "process_entry_id": process_entry_id,
                "break_prior_links": break_prior_links,
            }
            # Assuming make_api_request can handle the 'data' as it is
            new_link_response = self.client.make_api_request(
                "/components/instances/links", "PUT", data=data
            )

        # Clear the queue once all links have been created
        self.link_data_queue.clear()

    def submit(self, cycle_time=None, is_pass=None):
        """
        Mark a process entry as completed

        Args: 
        - cycle_time?: Optional number for seconds elapsed since the last cycle 
        finished for this process.
        - is_pass?: Optional boolean for indicating whether the process passed for
        this entry
        this entry

        Returns:
        - API response for submitting a process entry
        """
        data = {}
        if cycle_time is not None:
            data["cycle_time"] = cycle_time
        if is_pass is not None: 
            data["is_pass"] = is_pass
        try:
            self._process_text_data_queue()
            self._process_numerical_data_queue()
            self._process_file_data_queue()
            self._process_boolean_data_queue()
            self._process_link_data_queue()
        except SerialAPIException as e:
            raise SerialAPIException(f"Could not add data to process entry: {e}")

        data["is_complete"] = True

        self.data = self.client.make_api_request(f"/processes/entries/{self.id}", "PATCH", data=data)
        return self.data

class ProcessEntries:
    """
    A class for process entry data methods
    """
    @staticmethod
    def get(id):
        """
        Gets a process entry, if it exists
        
        Args:
        - id: Process entry id

        Returns:
        - A process entry Python object
        """
        client = APIClient()
        entry = client.make_api_request("/processes/entries", "GET", params={"id":id})
        if len(entry) == 0:
            raise SerialAPIException(f"Could not find process entry with id: {id}")
        entry = entry[0]
        return ProcessEntry(entry)

    @staticmethod
    def create(process_id, component_instance=None, component_instance_id=None, component_instance_identifier=None, station_id=None):
        """
        Creates a process entry

        Args:
        - process_id: Process id 
        - component_instance?: Component instance object 
        - component_instance_id?: Component instance id (is overriden by component_instance)
        - component_instance_identifier?: Component instance identifier (is overriden by component_instance & component_instance_id)
        - station_id?: Optional station id to override the default station id

        Returns:
        - A process entry Python object
        """
        client = APIClient()
        if component_instance:
            component_instance_id = component_instance.data["id"]
        if not component_instance_id and not component_instance_identifier:
            raise Exception("ComponentInstance id cannot be null, please pass in a valid one")
        if component_instance_identifier:
            component_instance_id = ComponentInstances.get(component_instance_identifier).data["id"]
        data = {"component_instance_id": component_instance_id, "process_id": process_id}
        if station_id:
            data["station_id"] = station_id
        elif config.station_id:
            data["station_id"] = config.station_id
        entry = client.make_api_request("/processes/entries", "POST", data=data)
        return ProcessEntry(entry)

    @staticmethod
    def list(query_params):
        """
        Lists process entries

        Args:
        - query_params: Query parameters for filtering process entries as defined at
        https://docs.serial.io/api-reference/process-entries/get-process-entries

        Returns:
        - A list of process entry Python objects, as defined above 
        """
        client = APIClient()
        entries = client.make_api_request("/processes/entries", "GET", params=query_params)
        entry_object_list = [ProcessEntry(entry) for entry in entries]
        return entry_object_list




from setuptools import setup, find_packages

setup(
    name="serialmfg",
    version="1.1.11",
    packages=find_packages(),
    description="This module provides a simple Python interface to interact with the Serial API. It allows you to easily perform common operations such as uploading process data, initializing identifiers, and checking the server connection.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/serialmfg/serial-py",
    author="Devon Copeland",
    author_email="founders@serial.io",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
    ],
    install_requires=["requests"],
    include_package_data=True
)



import sys
from dotenv import find_dotenv, load_dotenv
import os
sys.path.append('../serial')
import time
import serialmfg as serial
from .constants import NEW_IDENTIFIER_SAMPLE_DATA, SAMPLE_NEW_LINK, EXISTING_IDENTIFIER_DATA, SAMPLE_TEXT_DATA, SAMPLE_NUMBER_DATA, SAMPLE_IMAGE_DATA, SAMPLE_BOOLEAN_DATA, SAMPLE_ENTRY_DATA, SAMPLE_DATASET_DATA, SAMPLE_FILE_DATA

load_dotenv(find_dotenv()) # relative path to .env file
API_KEY = os.getenv('ZUPLO_API_KEY') # put your own API key here
BASE_URL = os.getenv('BASE_URL') # Put your own url here 
EXISTING_IDENTIFIER = "test-1691540942131"
UNEDITED_IDENTIFIER = "test-1696629404"

serial.set_api_key(API_KEY)
serial.set_base_url(BASE_URL)
serial.set_station_id("2f02d2be-ece6-410f-ad80-ffd746988870")

existing_component_instance = serial.ComponentInstances.get(EXISTING_IDENTIFIER) 
new_component_instance = serial.ComponentInstances.create(f"test-{int(time.time())}", component_name="Test Process Upload Component")

new_process_entry = serial.ProcessEntries.create(process_id="51718ea4-a274-4455-bde3-e4216e1ecd96", component_instance=existing_component_instance) 

existing_process_entry = serial.ProcessEntries.get('114b846e-5d5e-4f96-87d2-6c029192053a')

def test_get_instance():
    for key, value in existing_component_instance.data.items():
        assert key in EXISTING_IDENTIFIER_DATA.keys()
        if key not in ["id", "status", "identifier", "created_at", "last_updated_at", "completed_at"]:
            assert value == EXISTING_IDENTIFIER_DATA[key]

def test_create_instance():
    for key, value in new_component_instance.data.items():
        assert key in NEW_IDENTIFIER_SAMPLE_DATA.keys()
        if key not in ["id", "identifier", "created_at", "last_updated_at"]:
            assert value == NEW_IDENTIFIER_SAMPLE_DATA[key]
        
def test_list_instances():
    component_instance_list = serial.ComponentInstances.list({"status": "DEFECTIVE"})
    for instance in component_instance_list:
        assert instance.data["status"] == "DEFECTIVE"

class DummyProcessEntry: 
    def __init__(self):
        self.id = "1963ff32-cfb8-48a9-9a6d-41c7bfdcb67c"

def test_create_link():
    process_entry = DummyProcessEntry()
    new_link = new_component_instance.add_link("Test PyLibLinking dataset", UNEDITED_IDENTIFIER, process_entry=process_entry) 
    for key, value in new_link.data["new_link"].items():
        assert key in SAMPLE_NEW_LINK["new_link"].keys()
        if key not in ["id", "created_at", "removed_at", "unique_identifier_id"]:
            assert value == SAMPLE_NEW_LINK["new_link"][key]

def test_list_process_entries():
    process_entry_list = serial.ProcessEntries.list({"component_instance_id": "0372a807-b15f-47f9-acf9-d9a3301673ec"}) 
    for entry in process_entry_list:
        assert entry.data["unique_identifier_id"] == "0372a807-b15f-47f9-acf9-d9a3301673ec"

def test_get_process_entries():
    existing_process_entry = serial.ProcessEntries.get("e6ff9f24-8435-4ed7-9b4a-0f4b860ac1fc") 
    assert existing_process_entry.id == "e6ff9f24-8435-4ed7-9b4a-0f4b860ac1fc"

def test_create_process_entries():
    new_process_entry_2 = serial.ProcessEntries.create(process_id="51718ea4-a274-4455-bde3-e4216e1ecd96", component_instance=existing_component_instance) 

    assert new_process_entry_2.process_id == "51718ea4-a274-4455-bde3-e4216e1ecd96"
    assert new_process_entry_2.data["unique_identifier_id"] == "95db48e1-99ad-4e35-a86b-fa0beca5f313"

def test_create_process_entries_with_identifier():
    new_process_entry_2 = serial.ProcessEntries.create(process_id="51718ea4-a274-4455-bde3-e4216e1ecd96", component_instance_identifier=EXISTING_IDENTIFIER) 

    assert new_process_entry_2.process_id == "51718ea4-a274-4455-bde3-e4216e1ecd96"
    assert new_process_entry_2.data["unique_identifier_id"] == "95db48e1-99ad-4e35-a86b-fa0beca5f313"

def test_create_process_entries_with_id_and_submit():
    new_process_entry_2 = serial.ProcessEntries.create(process_id="51718ea4-a274-4455-bde3-e4216e1ecd96", component_instance_id="95db48e1-99ad-4e35-a86b-fa0beca5f313") 

    assert new_process_entry_2.process_id == "51718ea4-a274-4455-bde3-e4216e1ecd96"
    assert new_process_entry_2.data["unique_identifier_id"] == "95db48e1-99ad-4e35-a86b-fa0beca5f313"

    existing_process_entry.add_text("New Dataset", "Bob's Burgers") 
    existing_process_entry.add_number("New Dataset", 1.5, usl=5, lsl=0) 

    existing_process_entry.add_file("New Dataset", "/Users/clarke/repos/serial-py/tests/test.txt")

    existing_process_entry.add_boolean("Pass Fail Criteria", True, False)
    existing_process_entry.submit(cycle_time=50, is_pass=True)

    assert existing_process_entry.data["cycle_time"] == 50

def test_create_process_entries_with_identifier_and_fail():
    new_process_entry_2 = serial.ProcessEntries.create(process_id="51718ea4-a274-4455-bde3-e4216e1ecd96", component_instance_identifier=EXISTING_IDENTIFIER)

def test_get_datasets():
    dataset = serial.Datasets.get("LSL Only", "NUMERICAL", process_id="9ecb9747-22d9-49eb-b5c1-1d9fff3fa6b8")
    for key, value in dataset.data.items():
        assert key in SAMPLE_DATASET_DATA.keys()

def test_create_datasets():
    dataset = serial.Datasets.create("New Dataset", "NUMERICAL", "9ecb9747-22d9-49eb-b5c1-1d9fff3fa6b8", {"lsl": -10.5})
    for key, value in dataset.data.items():
        assert key in SAMPLE_DATASET_DATA.keys()




"""
This file contains the ComponentInstance class, which represents a component instance
"""
from ..api_client import APIClient
from ..exceptions import SerialAPIException
from .component_instance_link import ComponentInstanceLink

class ComponentInstance:
    """
    A component instance object, as defined at
    https://docs.serial.io/api-reference/component-instances/get-component-instance
    """
    def __init__(self, component_instance_data):
        """
        Args:
        - component_instance_data: A component instance object, as defined at
        https://docs.serial.io/api-reference/component-instances/get-component-instance
        
        Returns:
        - A component instance python object, which holds the api object at data,
        its links created during this session at created_links, and the client
        """
        self.client = APIClient()
        self.data = component_instance_data
        self.created_links = [] 

    def add_link(self, link_name, child_identifier, break_prior_links=False, process_entry=None):
        """
        Creates a link between this component instance and a child at a specific
        process_entry

        Args: 
        - link_name: User facing name for the link
        - child_identifier: Identifier for the child component instance to be linked
        - break_prior_links?: Boolean for whether to break prior links
        - process_entry?: Override for the process entry this component instance object 
        was created with. Must be included if this component instance object was not 
        created with a process entry.

        Returns:
        - New component instance link
        """

        # TODO: debug logging
        #print(f"Adding link: {link_name} with child identifier: {child_identifier}")
        child_component_instance_params = {
                "identifier": child_identifier
                }
        link_params = {
                "name": link_name
                }
        child_component_instance = self.client.make_api_request("/components/instances",
                                                                "GET",
                                                                params=child_component_instance_params)
        if len(child_component_instance) == 0:
            raise SerialAPIException(f"Component instance with identifier {child_identifier} does not exist")
        child_component_instance = child_component_instance[0]
        link_dataset = self.client.make_api_request("/datasets",
                                                    "GET",
                                                    params=link_params)
        if len(link_dataset) == 0:
            raise SerialAPIException(f"Dataset with name {link_name} does not exist")
        link_dataset = link_dataset[0]

        process_entry_id = None
        if process_entry:
            process_entry_id = process_entry.id
        else:
            process_entry_id = self.process_entry.id
        data = {
                "parent_component_instance_id": self.data["id"],
                "child_component_instance_id": child_component_instance["id"],
                "dataset_id": link_dataset["id"],
                "process_entry_id": process_entry_id,
                "break_prior_links": break_prior_links,
                }
        new_link = ComponentInstanceLink(self.client.make_api_request("/components/instances/links", "PUT", data=data)) 
        self.created_links.append(new_link)
        return new_link

class ComponentInstances:
    """
    A class for component instance data methods
    """
    @staticmethod
    def get(identifier):
        """
        Gets a component instance, if it exists
        
        Args:
        - identifier: Component's user facing identifier

        Returns:
        - A component instance object 
        """
        client = APIClient() 
        # TODO: debug logging
        #print(f"Getting component instance: {identifier}")
        return ComponentInstance(client.make_api_request(f"/components/instances/{identifier}", "GET"))
    
    @staticmethod
    def create(identifier, component_name):
        """
        Creates a component instance
        
        Args:
        - identifier: Component's user facing identifier
        - component_name: User facing name for the component
            
        Returns:
        - A component instance, as defined at 
        https://docs.serial.io/api-reference/component-instances/get-component-instance
        """
        # TODO: debug logging
        #print(f"Creating component instance: {identifier} with component name: {component_name}")
        client = APIClient() 
        # Get the component id for the given component name
        component_name_params = {
                "name" : component_name,
                }
        component_id = client.make_api_request("/components",
                                                     "GET",
                                                     params=component_name_params)
        if len(component_id) == 0:
            raise SerialAPIException(f"Component with name {component_name} does not exist")
        component_id = component_id[0]["id"]
        data = {
                "component_id": component_id,
                "identifier": identifier,
                }
        return ComponentInstance(client.make_api_request(f"/components/instances", "PUT", data=data)) 

    @staticmethod
    def list(query_params):
        """
        Lists component instances

        Args:
        - query_params: A dictionary of query parameters, as defined at
        https://docs.serial.io/api-reference/component-instances/get-component-instance

        Returns:
        - A list of component instances, as defined at
        https://docs.serial.io/api-reference/component-instances/get-component-instance
        """
        # TODO: debug logging
        #print(f"Listing component instances with query params: {query_params}")

        client = APIClient()
        instances = client.make_api_request("/components/instances", "GET", params=query_params)
        instance_object_list = [ComponentInstance(instance) for instance in instances]
        return instance_object_list 



"""
DEPRECATED, WILL BE REMOVED IN FUTURE RELEASE
"""
from warnings import warn 
class Identifier: 
    def __init__(self, identifier, component, part_number=None):
        """
        Initializes a new Identifier object.

        Args:
        - identifier: The identifier for the object.
        - component: The component for the object.
        - part_number: The part number for the object.

        Returns:
        - A new Identifier object.
        """
        self.identifier = identifier
        self.component = component
        self.part_number = part_number 
        warn('''
        The "Identifier" object is being deprecated and will be removed in future releases.
        Please begin migrating to the new interface provided from `from serialmfg import client`.
        ''')

